/* This class represents an input stream which repeats itself
   some number of times, or indefinitely. A non-positive repetition
   count means "indefinitely". */

package cgmud.audio;

import java.io.IOException;
import java.io.InputStream;


public class RepeatedInputStream extends InputStream {

    private byte[] d_streamData;
    private int d_pos;
    private int d_rpt;
    private int d_markpos, d_markrpt;


    /* ----- constructors ----- */

    public RepeatedInputStream(InputStream strm, int repeat)
    throws IOException
    {
	m_load(strm);
	strm.close();
	d_pos = 0;
	d_rpt = repeat;
	if (d_rpt == 0) d_rpt = -1;
	d_markpos = d_pos;
	d_markrpt = d_rpt;
    }

    private void m_load(InputStream fis)
    throws IOException
    {
	int datasize = 0;
	int bufsize = 65536;
	byte[] buf = new byte[bufsize];
	for ( ; ; ) {
	    int got = fis.read(buf, datasize, bufsize - datasize);
	    if (got == -1) break;
	    datasize += got;
	    if (datasize >= bufsize) {
		int newbs = bufsize * 2;
		byte[] newbuf = new byte[newbs];
		System.arraycopy(buf, 0, newbuf, 0, datasize);
		bufsize = newbs;
		buf = newbuf;
	    }
	}
	if (datasize == bufsize) {
	    d_streamData = buf;
	} else {
	    d_streamData = new byte[datasize];
	    System.arraycopy(buf, 0, d_streamData, 0, datasize);
	}
    }


    /* ----- InputStream overrides ----- */

    private boolean m_atEOF() {
	if (d_pos < d_streamData.length) return false;
	if (d_rpt > 0) d_rpt -= 1;
	if (d_rpt == 0) return true;
	d_pos = 0;
	return false;
    }

    public int read() {
	if (m_atEOF()) return -1;
	return (int) d_streamData[d_pos++];
    }

    public int read(byte[] buf, int offset, int len) {
	if (len < 0) return 0;
	int amt = available();
	if (amt <= 0) return -1;
	if (amt > len) amt = len;
	if (buf != null) {
	    System.arraycopy(d_streamData, d_pos, buf, offset, amt);
	}
	d_pos += amt;
	return amt;
    }

    public long skip(long len) {
	if (len < 0L) return 0L;
	int amt = available();
	if (len < (long) amt) {
	    amt = (int) len;
	}
	d_pos += (int) len;
	return (long) amt;
    }

    public int available() {
	if (m_atEOF()) return 0;
	return d_streamData.length - d_pos;
    }

    public synchronized void mark(int readlimit) {
	d_markpos = d_pos;
	d_markrpt = d_rpt;
    }

    public synchronized void reset() {
	d_pos = d_markpos;
	d_rpt = d_markrpt;
    }

    public boolean markSupported() { return true; }

}
