/* Copyright 1999 by Donald Reble. All rights reserved. */

/* $Id: $ */
/* This class represents a sequence of bytes.
   Multi-byte integers are in network-byte order (big endian).

   One must explicitly allocate space for data. Attempting to append
   past the capacity is an error.
*/

package cgmud.base;


public class ByteSequence {

    private byte[] d_data;
    private int d_index;
    private int d_allocsize;


    /* ----- constructors ----- */

    public ByteSequence(int size) {
	d_allocsize = size;
	d_data = new byte[d_allocsize];
	d_index = 0;
    }

    public ByteSequence(ByteSequence bs) {
	d_allocsize = bs.d_data.length;
	d_data = new byte[d_allocsize];
	d_index = 0;
	System.arraycopy(bs.d_data, 0, d_data, 0, d_allocsize);
    }


    /* ----- inspectors ----- */

    public int getIndex() { return d_index; }
    public int getSize() { return d_allocsize; }
    public boolean atEnd() { return d_index >= d_allocsize; }

    public byte peekByte() { return d_data[d_index]; }

    public int m_getByte() {
	int val = ((int) d_data[d_index]) & 0x000000ff;
	d_index += 1;
	return val;
    }

    public byte getByte() {
	byte val = d_data[d_index];
	d_index += 1;
	return val;
    }

    public short getShort() {
	int val = m_getByte() << 8;
	val |= m_getByte();
	return (short) val;
    }

    public int getInt() {
	int val = m_getByte() << 24;
	val |= m_getByte() << 16;
	val |= m_getByte() << 8;
	val |= m_getByte();
	return val;
    }

    public boolean getBoolean() {
	return getByte() != 0;
    }

    public byte[] getBlock(int length) {
	byte[] blk = new byte[length];
	System.arraycopy(d_data, d_index, blk, 0, length);
	d_index += length;
	return blk;
    }

    public String getString() {
	int sz = d_data.length;
	StringBuffer buf = new StringBuffer();
	while ((d_index < sz) && (d_data[d_index] != 0)) {
	    buf.append((char) d_data[d_index]);
	    d_index += 1;
	}
	if (d_index < sz) d_index += 1; // past trailing null
	return buf.toString();
    }


    /* ----- mutators ----- */

    public byte[] getArray() { return d_data; }

    public void expand(int size) {
	if (size > d_allocsize) {
	    byte[] data = new byte[size];
	    System.arraycopy(d_data, 0, data, 0, d_allocsize);
	    d_data = data;
	    d_allocsize = size;
	}
    }

    public void shrink() {
	if (d_allocsize > d_index) {
	    byte[] data = new byte[d_index];
	    System.arraycopy(d_data, 0, data, 0, d_index);
	    d_data = data;
	    d_allocsize = d_index;
	}
    }


    public void setIndex(int index) {
	d_index = index;
    }


    public void appendByte(byte data) {
	d_data[d_index] = data;
	d_index += 1;
    }

    public void appendShort(short data) {
	d_data[d_index] = (byte) (data >> 8);
	d_index += 1;
	d_data[d_index] = (byte) (data);
	d_index += 1;
    }

    public void appendInt(int data) {
	d_data[d_index] = (byte) (data >> 24);
	d_index += 1;
	d_data[d_index] = (byte) (data >> 16);
	d_index += 1;
	d_data[d_index] = (byte) (data >> 8);
	d_index += 1;
	d_data[d_index] = (byte) data;
	d_index += 1;
    }

    public void appendBoolean(boolean data) {
	appendByte((byte) (data ? 1 : 0));
    }

    public void appendBlock(int length, byte[] data) {
	for (int ix = 0; ix < length; ix += 1) {
	    d_data[d_index] = data[ix];
	    d_index += 1;
	}
    }

    public void appendString(String str) {
	int sz = str.length();
	for (int ix = 0; ix < sz; ix += 1) {
	    d_data[d_index] = (byte) str.charAt(ix);
	    d_index += 1;
	}
	d_data[d_index] = '\0';
	d_index += 1;
    }

}
