/* Copyright 1999 by Donald Reble. All rights reserved. */

/* $Id: $ */
/* This class represents a client-server message.
   Multi-byte integers are in network-byte order.
   Format:
	bytes 0..1: tail length
	bytes 2..5: key
	bytes 6:    kind
	bytes 7:    flag
	bytes 8+    someint
	bytes 10+   tail: arbitrary data

    One should extend this class for each actual message in use.
*/

package cgmud.message;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import cgmud.base.FailureException;
import cgmud.base.ByteSequence;
import cgmud.main.LogFile;
import cgmud.main.Globals;
import cgmud.main.Main;


public abstract class Message extends ByteSequence {

    public static final byte
	      KILL_CLIENT =  1,
	      NUKE_CLIENT =  2,
	       RUN_CLIENT =  3,
	     TEXT_MESSAGE =  4,
	ENTER_WIZARD_MODE =  5,
	LEAVE_WIZARD_MODE =  6,
	     SET_CONTINUE =  7,
	  CHANGE_PASSWORD =  8,
	       SET_PROMPT =  9,
	      DELETE_PROC = 10,
		  EFFECTS = 11,
	      EDIT_STRING = 12,
		EDIT_PROC = 13,
	       GET_STRING = 14,
	     FLUSH_SYMBOL = 15,
	    DEFINE_EFFECT = 16,
	       QUERY_FILE = 17,
		SHUT_DOWN = 18,
		    FLUSH = 19,
	   SYNC_SHUT_DOWN = 20,
	     BEGIN_CLIENT = 21,
	       END_CLIENT = 22,
	       ASK_PLAYER = 23,
	   CREATION_CHECK = 24,
	    CREATE_PLAYER = 25,
	     LOGIN_PLAYER = 26,
		   LOOKUP = 27,
		FIND_NAME = 28,
		READ_PROC = 29,
	   WRITE_PROC_NEW = 30,
		 RUN_PROC = 31,
		      LOG = 32,
	       INPUT_LINE = 33,
		 SET_SYNC = 34,
		  RAW_KEY = 35,
	    REGION_SELECT = 36,
	       BUTTON_HIT = 37,
	     FLUSH_EFFECT = 38,
	      TEXT_RESIZE = 39,
		  INC_REF = 40,
		  DEC_REF = 41,
	     SYMBOL_ENTER = 42,
	    SYMBOL_DELETE = 43,
		USE_TABLE = 44,
	      UNUSE_TABLE = 45,
	 DESCRIBE_BUILTIN = 46,
		 DESCRIBE = 47,
	    GRAPHICS_FLIP = 48,
	       VOICE_FLIP = 49,
	       SOUND_FLIP = 50,
	       MUSIC_FLIP = 51,
	     DO_EDIT_PROC = 52,
	 EDIT_STRING_DONE = 53,
	     REPLACE_PROC = 54,
	   EDIT_PROC_DONE = 55,
	  GET_STRING_DONE = 56,
	      EFFECT_DONE = 57,
	  QUERY_FILE_DONE = 58,
	 CREATE_CONTAINER = 59,
	 CREATE_COMPONENT = 60,
	       MAKE_FRAME = 61;


    private   static final short D_LOC_TAILLEN  =  0;
    private   static final short D_LOC_KEY      =  2;
    private   static final short D_LOC_KIND     =  6;
    private   static final short D_LOC_FLAG     =  7;
    protected static final short D_LOC_PARMINT  =  8;
    protected static final short D_LOC_TAIL     = 10;
    public    static final short PASSWORD_LEN = 20;
    protected static final short D_LOC_AFTER_PW = D_LOC_TAIL + PASSWORD_LEN;

    protected static byte[] d_sessionKey = new byte[4];


    /* ----- class methods ----- */

    public static void setSessionKey(int key) {
	d_sessionKey[0] = (byte) (key >> 24);
	d_sessionKey[1] = (byte) (key >> 16);
	d_sessionKey[2] = (byte) (key >>  8);
	d_sessionKey[3] = (byte) key;
    }


    /* ----- constructors ----- */

    protected Message(byte kind, int key, byte flag, short someint,
		      short taillen)
    {
	super(D_LOC_TAIL + taillen);
	appendShort(taillen);
	appendInt(key);
	appendByte(kind);
	appendByte(flag);
	appendShort(someint);
    }

    protected Message(byte kind, int key, byte flag, short taillen) {
	this(kind, key, flag, (short) 0, taillen);
    }

    protected Message(ByteSequence bs) { super(bs); }


    /* ----- I/O ----- */

    public void send(BufferedOutputStream bs)
    throws IOException
    {
	bs.write(getArray(), 0, getSize());
	bs.flush();
    }

    private static int m_get(BufferedInputStream strm,
			     byte[] bs, int idx, int len)
    throws IOException
    {
	int amt = 0;
	while (len > 0) {
	    int sz = strm.read(bs, idx, len);
	    if (sz <= 0) break;
	    idx += sz;
	    amt += sz;
	    len -= sz;
	}
	return amt;
    }

    /* returns null when EOF encountered */
    public static Message receive(BufferedInputStream strm)
    throws IOException
    {
	ByteSequence bs = new ByteSequence(D_LOC_TAIL);
	int sz = m_get(strm, bs.getArray(), 0, D_LOC_TAIL);
	if (sz < D_LOC_TAIL) {
	    if (sz != 0) {
		Globals.logFile.logStringAlways(
		    "Message.receive got only " + sz + " in header."
		);
		Globals.logFile.logExceptional(bs);
	    }
	    return null; // EOF, I guess.
	}

	bs.setIndex(D_LOC_TAILLEN);
	int len = ((int) bs.getShort()) & 0x0000ffff;
	if (len > 0) {
	    bs.expand(len + D_LOC_TAIL);
	    sz = m_get(strm, bs.getArray(), D_LOC_TAIL, len);
	    if (sz < len) {
		Globals.logFile.logStringAlways(
		    "Message.receive wanted " + len + "; got only " +
		    sz + " in header.");
		Globals.logFile.logExceptional(bs);
		return null; // Treat as EOF.
	    }
	}
	bs.setIndex(D_LOC_KIND);
	switch (bs.getByte()) {
	    case KILL_CLIENT:	    return new KillClient(bs);
	    case NUKE_CLIENT:	    return new NukeClient(bs);
	    case RUN_CLIENT:	    return new RunClient(bs);
	    case TEXT_MESSAGE:	    return new TextMessage(bs);
//	    case ENTER_WIZARD_MODE: return new EnterWizardMode(bs);
//	    case LEAVE_WIZARD_MODE: return new LeaveWizardMode(bs);
//	    case SET_CONTINUE:	    return new SetContinue(bs);
	    case CHANGE_PASSWORD:   return new ChangePassword(bs);
	    case SET_PROMPT:	    return new SetPrompt(bs);
//	    case DELETE_PROC:	    return new DeleteProc(bs);
	    case EFFECTS:	    return new Effects(bs);
//	    case EDIT_STRING:	    return new EditString(bs);
//	    case EDIT_PROC:	    return new EditProc(bs);
	    case GET_STRING:	    return new GetString(bs);
//	    case FLUSH_SYMBOL:	    return new FlushSymbol(bs);
	    case DEFINE_EFFECT:     return new DefineEffect(bs);
//	    case QUERY_FILE:	    return new QueryFile(bs);
//	    case SHUT_DOWN:	    return new ShutDown(bs);
//	    case FLUSH:		    return new Flush(bs);
//	    case SYNC_SHUT_DOWN:    return new SyncShutDown(bs);
	    case BEGIN_CLIENT:	    return new BeginClient(bs);
	    case END_CLIENT:	    return new EndClient(bs);
	    case ASK_PLAYER:	    return new AskPlayer(bs);
	    case CREATION_CHECK:    return new CreationCheck(bs);
	    case CREATE_PLAYER:     return new CreatePlayer(bs);
	    case LOGIN_PLAYER:      return new LoginPlayer(bs);
//	    case LOOKUP:	    return new Lookup(bs);
//	    case FIND_NAME:	    return new FindName(bs);
//	    case READ_PROC:	    return new ReadProc(bs);
//	    case WRITE_PROC_NEW:    return new WriteProcNew(bs);
//	    case RUN_PROC:	    return new RunProc(bs);
	    case LOG:		    return new Log(bs);
	    case INPUT_LINE:	    return new InputLine(bs);
//	    case SET_SYNC:	    return new SetSync(bs);
	    case RAW_KEY:	    return new RawKey(bs);
	    case REGION_SELECT:     return new RegionSelect(bs);
	    case BUTTON_HIT:	    return new ButtonHit(bs);
//	    case FLUSH_EFFECT:	    return new FlushEffect(bs);
//	    case TEXT_RESIZE:	    return new TextResize(bs);
//	    case INC_REF:	    return new IncRef(bs);
//	    case DEC_REF:	    return new DecRef(bs);
//	    case SYMBOL_ENTER:	    return new SymbolEnter(bs);
//	    case SYMBOL_DELETE:     return new SymbolDelete(bs);
//	    case USE_TABLE:	    return new UseTable(bs);
//	    case UNUSE_TABLE:	    return new UnuseTable(bs);
//	    case DESCRIBE_BUILTIN:  return new DescribeBuiltin(bs);
//	    case DESCRIBE:	    return new Describe(bs);
//	    case GRAPHICS_FLIP:     return new GraphicsFlip(bs);
//	    case VOICE_FLIP:	    return new VoiceFlip(bs);
//	    case SOUND_FLIP:	    return new SoundFlip(bs);
//	    case MUSIC_FLIP:	    return new MusicFlip(bs);
//	    case DO_EDIT_PROC:	    return new DoEditProc(bs);
//	    case EDIT_STRING_DONE:  return new EditStringDone(bs);
//	    case REPLACE_PROC:	    return new ReplaceProc(bs);
//	    case EDIT_PROC_DONE:    return new EditProcDone(bs);
	    case GET_STRING_DONE:   return new GetStringDone(bs);
//	    case EFFECT_DONE:	    return new EffectDone(bs);
//	    case QUERY_FILE_DONE:   return new QueryFileDone(bs);
	    case CREATE_CONTAINER:  return new CreateContainer(bs);
	    case CREATE_COMPONENT:  return new CreateComponent(bs);
	    case MAKE_FRAME:        return new MakeFrame(bs);
	    default: return new IgnoreMessage(bs); /* %%% stopgap */
	}
    }


    /* ----- inspectors ----- */

    public short getTailLength() {
	setIndex(D_LOC_TAILLEN);
	return getShort();
    }

    public int getKey() {
	setIndex(D_LOC_KEY);
	return getInt();
    }

    public byte getKind() {
	setIndex(D_LOC_KIND);
	return getByte();
    }

    public boolean getFlag() {
	setIndex(D_LOC_FLAG);
	return getByte() != 0;
    }

    public short getParmint() {
	setIndex(D_LOC_PARMINT);
	return getShort();
    }

    public byte[] getPassword() {
	setIndex(D_LOC_TAIL);
	return getBlock(PASSWORD_LEN);
    }


    /* ----- processing ----- */

    public abstract void process(Main main);

}
